# Generated by Django 3.1.1 on 2021-10-08 10:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0009_set_old_tokens_to_sso_v1'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='sso_version',
            field=models.IntegerField(default=2, help_text='EVE SSO Version.'),
        ),
    ]
