# Generated by Django 3.1.3 on 2020-11-09 23:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0007_fix_mysql_8_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='refresh_token',
            field=models.TextField(editable=False, help_text='A re-usable token to generate new access tokens upon expiry.', null=True),
        ),
    ]
