# Generated by AaronKable on 2020-10-03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0006_remove_url_length_limit'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='refresh_token',
            field=models.TextField(editable=False, help_text='A re-usable token to generate new access tokens upon expiry.'),
        ),
    ]
