# Generated by Django 3.1.1 on 2021-10-08 10:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0008_nullable_refresh_token'),
    ]

    operations = [
        migrations.AddField(
            model_name='token',
            name='sso_version',
            field=models.IntegerField(default=1, help_text='EVE SSO Version.'),
        ),
    ]
